/*
 * PalmOSFileIO.c
 * Procedures to handle file IO on Palm
 */
#define _PALMOSFILEIO_C

#include <PalmOS.h>
#include <MemoryMgr.h>
#include <DataMgr.h>

#include "PalmOSGUI.h"
#include "PalmOSArmlet.h"
#include "PalmOSFileIO.h"
#include "FrodoResources.h"
#include "PalmHelper.h"
#include "PalmHardware.h"


// lastName is set on succesfull load and save, and is used to suggest a file name on 'Save As'
// it is also used as last snapshot name for 'Save to last'  and 'Load last'
#define INVALID_SNAPSHOT_INDEX -1
// as we ignore index 0 = default snapshot, we may change invalid index value to 0 and index type back to UInt16
Int32 lastIndex = INVALID_SNAPSHOT_INDEX;
char lastName[32] = "Unnamed";
UInt16 lastCard = 0xffff;
char *frodoDir = "/Palm/programs/Frodo";

// The disk/snapshot database
LocalID mainDB;				// ID of main database
DmOpenRef mainDBRef = 0;	// Open reference to main database
FileRef mainFileRef = 0;  // Open reference to Frodo.prc on SD card

// The mounted disk (8)
FILE disk;

// Structures to handle list of files
UInt16 nfile = 0, maxfile = 0;
MemHandle fileH = NULL;
struct t_file {
	UInt16 card;   // 0xffff if internal pdb
	char name[32]; // file name in /Palm/programs/Frodo/
} *file = NULL;

UInt16 nitem = 0, maxitem = 0;
MemHandle itemH = NULL;
struct t_item {
//	struct t_file *file;
  UInt16 fileindex;
	UInt16 index; // 0xffff if not a pdb
	char name[34];
} *item = NULL;

UInt16 nitemList = 0, maxitemList = 0;
MemHandle itemListH = NULL;
char **itemList = NULL;
Boolean bSyncRequired = false;


//***********************************************************
// Procedure: InitDB
//
// Parameter: -
//
// Returnval: -
//
// Description: Initialize main frodo database
//***********************************************************
#define DB_NAME "Frodo main database"
void InitDB(void)
{
	Err e;
	UInt16 attr;
	UInt16 at;
	UInt16 card;
	LocalID id;
	DmOpenRef dataDBRef;
	Boolean newDB = false;
	MemHandle h;
	MemPtr p;
	DmSearchStateType search;
	UInt16 i;
	
	// Check if Sync is required after each write operation
	bSyncRequired = DeviceHasNVFS();
	
  if(!bLaunchedFromSD)
  {
    //-----------------------------------------------
    // Initialize internal main database
    //-----------------------------------------------
  	mainDB = DmFindDatabase(0, DB_NAME);
  	if (!mainDB)
  	{
  		e = DmCreateDatabase(0, DB_NAME, appSignature, mainType, false);
  		if (e) 
  		{
  			ShowErrorResLong(StringErrFailedToCreateDB, e);
  			return;
  		}
  
  		newDB = true;
    	mainDB = DmFindDatabase(0, DB_NAME);
    	if (!mainDB) 
    	{
    		ShowErrorRes(StringErrSilentCreateDB);
    		return;
    	}
  	}
  	attr = dmHdrAttrBackup | dmHdrAttrBundle;
  	DmSetDatabaseInfo(0, mainDB, NULL, &attr, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

  	mainDBRef = DmOpenDatabase(0, mainDB, dmModeReadWrite);
  	if (!mainDBRef) 
  	{
  		ShowErrorRes(StringErrFailedToOpenDB);
  		return;
  	}
  	if (newDB) 
  	{
  		// Create entry for snapshot in new created database
  		at = 0;
  		h = DmNewRecord(mainDBRef, &at, SNAPSHOT_SIZE_1+36);
  		if (!h) 
  		{
  			ShowErrorResLong(StringErrFailedToCreateSnap, DmGetLastErr());
  			DmCloseDatabase(mainDBRef);
  			DmDeleteDatabase(0, mainDB);
  			mainDB = 0;
  			mainDBRef = NULL;
  			return;
  		}
  		p = MemHandleLock(h);
  		e = DmWrite(p, 0, "SNAP", 4);
  		if(!e)
  			e = DmWrite(p, 4, "Nuked snapshot\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 32);
  		if(!e)
  			e = DmWrite(p, 36, "NUKE", 4); 
  		MemHandleUnlock(h);
  		DmReleaseRecord(mainDBRef, at, true);
  		if(e)
  		{
  			ShowErrorResLong(StringErrDmWriteFailed, e);
  			DmCloseDatabase(mainDBRef);
  			DmDeleteDatabase(0, mainDB);
  			mainDB = 0;
  			mainDBRef = NULL;
  			return;
  		}
  		at = 1;
  		h = DmNewRecord(mainDBRef, &at, SNAPSHOT_SIZE_2+36+0x30);
  		if (!h) 
  		{
  			ShowErrorResLong(StringErrFailedToCreateSnap, DmGetLastErr());
  			DmCloseDatabase(mainDBRef);
  			DmDeleteDatabase(0, mainDB);
  			mainDB=0;
  			mainDBRef=NULL;
  			return;
  		}
  		p = MemHandleLock(h);
  		e = DmWrite(p, 0, "SNP2", 4);
  		if(!e)
  			e = DmWrite(p, 4, "Nuked snapshot\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 32);
  		if(!e)
  			e = DmWrite(p, 36, "NUKE", 4); 
  		MemHandleUnlock(h);
  		DmReleaseRecord(mainDBRef, at, true);
  		if(e)
  		{
  			ShowErrorResLong(StringErrDmWriteFailed, e);
  			DmCloseDatabase(mainDBRef);
  			DmDeleteDatabase(0, mainDB);
  			mainDB = 0;
  			mainDBRef = NULL;
  			return;
  		}
  	}
  
  	// Database is open. Look for other Frodo PDB's and merge them into the main db.
  	if (DmGetNextDatabaseByTypeCreator(true, &search, dataType, appSignature, false, &card, &id) == errNone)
  	{
  		do 
  		{
  			UInt16 n;
  
  			dataDBRef = DmOpenDatabase(card, id, dmModeReadWrite);
  			if (!dataDBRef) 
  			{
  				ShowErrorRes(StringErrFailedToOpenSecDB);
  				continue;
  			}
  			n = DmNumRecords(dataDBRef);
  			for (i=0; i<n; i++) 
  			{
  				e = DmDetachRecord(dataDBRef, 0, &h);
  				if (e) 
  				{
  					ShowErrorResLong(StringErrDetachFailed, i);
  					continue;
  				}
  				at = dmMaxRecordIndex;
  				e = DmAttachRecord(mainDBRef, &at, h, NULL);
  				if (e) 
  				{
  					ShowErrorResLong(StringErrLostSomeData, e);
  				}
  			}
  			DmCloseDatabase(dataDBRef);
  			DmDeleteDatabase(card, id);
  		} while (DmGetNextDatabaseByTypeCreator(false, &search, dataType, appSignature, false, &card, &id) == errNone);
  	}
  }
  	
	// We'll do this everytime just in case
	VFSRegisterDefaultDirectory(".d64", expMediaType_Any, frodoDir);
	VFSRegisterDefaultDirectory(".t64", expMediaType_Any, frodoDir);
	VFSRegisterDefaultDirectory(".p00", expMediaType_Any, frodoDir);
	// Seems to be case insensitive, but just to be safe
	VFSRegisterDefaultDirectory(".D64", expMediaType_Any, frodoDir);
	VFSRegisterDefaultDirectory(".T64", expMediaType_Any, frodoDir);
	VFSRegisterDefaultDirectory(".P00", expMediaType_Any, frodoDir);
	// Stand-alone snapshots are not supported yet
	//VFSRegisterDefaultDirectory(".snp", expMediaType_Any, frodoDir);

	// If we crashed last time, the snapshot is not to be trusted. "Safe Mode"...
	if (!newDB && !crashed)
	{
		if(bLaunchedFromSD)
		{
		  e = LoadSnapshotFromSD(lastCard, DEFAULT_SNAP_NAME ".snp");
		  if(e)
  		{
		    // No snapshot -> start new
  			armletData->cmd &= ~(CMD68K_LOAD);
  		}
		}
		else
		{
  		Load(NULL, 0, 0xffff); // Load snapshot saved on last exit
    }
		if(prefs->DrivePath[0] != 0)
		  doFileLoadAfterSnap = true;
  }
  
	return;
}


//***********************************************************
// Procedure: FreeDB
//
// Parameter: -
//
// Returnval: -
//
// Description: Release main frodo database
//***********************************************************
void FreeDB(void)
{
	Close((FILE *) swap32((UInt32)armletData->disk));
	
	if(!bLaunchedFromSD)
  	DmCloseDatabase(mainDBRef);
  FreeItems();
}


//***********************************************************
// Procedure: Load
//
// Parameter: dbf   Pointer to file definition (see PalmOSGUI_dbfile.h)
//            id    ID (index) of record in pdb file
//            card  Card of file
//
// Returnval: -
//
// Description: Load a snapshot, d64 or t64 from the main db
//***********************************************************
void Load(DbFile *dbf, UInt16 id, UInt16 card)
{
	Err err;
	DbFile mainDbf;
	MemHandle h1 = NULL;
	MemHandle h2 = NULL;
	UInt8 *p1 = NULL;
	UInt8 *p2 = NULL;
	char name[32];
	Boolean ok = false;
	Boolean nuked = false;
	Boolean snap = false;
	UInt16 i;
	Int32 hit = -1;
  
	if (!dbf) 
	{
		// Load snapshot saved on last exit
		dbf = &mainDbf;
		DbFileFromDbRef(mainDBRef, dbf);
	}
	if (!dbf->u.db)
	  return;

	// Find by index
	err = DbFileGetRecord(dbf, id, &h1);
	if (h1) 
	{
		p1 = MemHandleLock(h1);
		if (p1) 
		{
			if (!MemCmp(p1, "SNAP", 4) || !MemCmp(p1, "SPLT", 4)) 
			{
				hit = id;
				if (!MemCmp(p1, "SNAP", 4))
					snap = true;
			} 
			else 
			{
				StrPrintF(name, "Invalid type \"%c%c%c%c\"", p1[0], p1[1], p1[2], p1[3]);
				ShowError(name);
				p1 = NULL;
				MemHandleUnlock(h1);
				DbFileReleaseRecord(dbf, id, h1, false);
			}
		} 
		else 
		{
			ShowErrorRes(StringErrFailedToLockHandle);
			DbFileReleaseRecord(dbf, id, h1, false);
		}
	}
	
	if (hit<0)
		return;
  
	// File found, and locked in p1.
	if (!MemCmp(p1, "SNAP", 4)) 
	{
		// Snapshots
		err = DbFileGetRecord(dbf, hit+1, &h2);
		if (h2) 
		{
			p2 = MemHandleLock(h2);
			if (p2) 
			{
				if ((!StrNCompare((char *) (p1+36), "NUKE", 4)) || (!StrNCompare((char *) (p2+36), "NUKE", 4)))
					nuked = true;
				else if (!MemCmp(p2, "SNP2", 4) && !StrNCompare(p1+4, p2+4, 32)) 
				{
		      // Copy snapshot data into memory
		      MemMove((MemPtr)swap32((UInt32)armletData->snapshot1), p1+36, SNAPSHOT_SIZE_1);
					MemMove((MemPtr)swap32((UInt32)armletData->snapshot2), p2+36, SNAPSHOT_SIZE_2);
  				if(MemHandleSize(h2) > 36 + SNAPSHOT_SIZE_2)
  				{
  				  // Snapshot with name of disk
    				MemMove(prefs->DrivePath, p2+36+SNAPSHOT_SIZE_2, 32);
  					MemMove(&(prefs->DriveCard), p2+36+SNAPSHOT_SIZE_2+32, 4);
  					MemMove(&(prefs->DriveIndex), p2+36+SNAPSHOT_SIZE_2+36, 4);
          }
					ok = true;
					if (hit != 0) // ignore default snapshot with index  0
					{ 
						lastIndex = hit;
						StrNCopy(lastName, p1+4, 31);
						lastName[31]='\0';
						lastCard = 0xffff;
					}
				}
				MemHandleUnlock(h2);
			}
			DbFileReleaseRecord(dbf, hit+1, h2, false);
		}
		MemHandleUnlock(h1);
		DbFileReleaseRecord(dbf, hit, h1, false);
		if (ok) 
		{
			armletData->cmd |= CMD68K_LOAD;
  		if(prefs->DrivePath[0] != 0)
  		  doFileLoadAfterSnap = true;
		} 
		else 
		{
			armletData->cmd &= ~(CMD68K_LOAD);
			if (!(nuked && id==0))
				ShowErrorRes(StringErrFailedToLoadSnap);
		}
	} 
	else if (!MemCmp(p1, "SPLT", 4)) 
	{
		// SplitFile .d64 or .t64

/* Current layout:
	00 'SPLT' (first part) || 'PART' (other parts)
	04 char name[32]     // must have terminating 0, so max 31 letters.
	36 int32 totalParts
	40 int32 part
	44 int32 size        // File size
	48 int32 partSize    // Part size. All parts must be the same size (except the last)
	52 char data[partSize]
*/
		int nParts;
		Close((FILE *) swap32((UInt32)armletData->disk));
		armletData->disk = NULL;
		MemSet(&disk, sizeof(FILE), 0x0);
		MemMove(disk.name, p1+4, 32);
		disk.part[0].p = p1;

		nParts = *((UInt32 *) (p1+36));
		disk.size = *((UInt32 *) (p1+44));
		disk.partSize = *((UInt32 *) (p1+48));
		disk.part[0].handle = (UInt32) h1;
    disk.modified = 0;
  	disk.part[0].dbf = (UInt32)dbf;
  	disk.part[0].id = hit;

		if (dbf->external) 
		{
			disk.location = 1;
		} 
		else 
		{
			disk.location = 0;
		}
		ok = true;
		
		for (i=0; i<nParts; i++) 
		{
			int n, p;
			UInt32 s, ps;
			if (i>0) 
			{
				// next part
				err = DbFileGetRecord(dbf, id+i, (MemHandle *) &(disk.part[i].handle));
				h1 = (MemHandle) disk.part[i].handle;
				if (err) 
				{
					ok = false;
					break;
				}
				disk.part[i].id = id+i;
					
				if (!(disk.part[i].p = p1 = MemHandleLock(h1))) 
				{
					ok = false;
					break;
				}
				
				if (MemCmp(p1, "PART", 4)) 
				{
					ok = false;
					break;
				}
			}
			n = *((UInt32 *) (p1+36)); // Total parts
			p = *((UInt32 *)(p1+40));  // Current part
			s = *((UInt32 *)(p1+44));  // Total size
			ps = *((UInt32 *)(p1+48)); // Size of current part
			
			if (n!=nParts || p!=i || s!=disk.size || ((i<p-1) && ps!=disk.partSize)) 
			{
				ok = false;
				break;
			}
			disk.part[i].p = p1+52;
			if (i == nParts-1) 
			{
				StrNCopy(lastName, p1+4, 32);
				lastName[31]='\0';
			}
		}
		if (ok) 
		{
			disk.size = swap32(disk.size);
			disk.partSize = swap32(disk.partSize);
			for (i=0; i<nParts; i++)
				disk.part[i].p = (UInt8 *) swap32((UInt32)disk.part[i].p);
			armletData->disk = (FILE *) swap32((UInt32) &disk);
			if (!StrNCaselessCompare(disk.name + StrLen(disk.name)-4, ".d64", 4))
				prefs->DriveType = swap32(DRVTYPE_D64);
			else
			{
				prefs->DriveType = swap32(DRVTYPE_T64);
				if(prefs->Emul1541Proc)
				{
				  // No emulation if t64 or p00 file
				  prefs->Emul1541Proc = 0;
				  ShowInfoRes(StringWarn1541Off);
					SavePrefs();
				  SavePrefsPalm();
				}
			}
			MemMove(prefs->DrivePath, disk.name, 32);
      prefs->DriveCard = card;
      prefs->DriveIndex = id;
      
			armletData->cmd |= CMD68K_SETPREFS;
		} 
		else 
		{
			armletData->cmd &= ~(CMD68K_SETPREFS);
			ShowErrorResStr(StringErrFailedToLoadDiskImage, disk.name);
		}
	} 
	else 
	{
		StrPrintF(name, "Unknown file type \"%c%c%c%c\"",	p1[0], p1[1], p1[2], p1[3]);
		ShowError(name);
		MemHandleUnlock(h1);
		DbFileReleaseRecord(dbf, hit, h1, false);
	}
}


//***********************************************************
// Procedure: Close
//
// Parameter: f   Definition of file(see Armlet68k.h)
//
// Returnval: -
//
// Description: Close file
//***********************************************************
void Close(FILE *f)
{
	int i;
  char strAsk[128];
  
	if (!f) 
		return;
	
  if(f->modified && f->location != 0 && prefsPalm->SaveToSDCard != SAVE2SD_NEVER)
  {
    if(prefsPalm->SaveToSDCard == SAVE2SD_ASK)
    {
      StrPrintF(strAsk, "Save '%s' back to SD card?", f->name);
      i = AskUserDialog(strAsk);
    }
    else
      i = 1; // SaveToSDCard == SAVE2SD_ALWAYS
    if(i)
      SaveFile2SD(f);
	}
	
	for (i=0; i<MAX_PARTS; i++) 
	{
		if (f->part[i].handle) 
		{
			if (f->part[i].p)
				MemHandleUnlock((MemHandle) f->part[i].handle);
			if (f->location == 0)
				DmReleaseRecord(mainDBRef, f->part[i].id, false);
			else
				MemHandleFree((MemHandle) f->part[i].handle);
		}
		f->part[i].handle = NULL;
		f->part[i].p = NULL;
		f->part[i].dbf = NULL;
	}
}


//***********************************************************
// Procedure: SaveFile2SD
//
// Parameter: f   Pointer to file definition (see Armlet68k.h)
//
// Returnval: -
//
// Description: Save file back to SD card. 
//***********************************************************
void SaveFile2SD(FILE *f)
{
  int i;
  Err err;
  FileRef fr = NULL;
  UInt16 lNumRecs;
  UInt32 lNumBytesIO;
  Int32 lRemainingSize;
  UInt32 lCurrPart;
  UInt32 lIndex;
  UInt32 lStartOffset, lEndOffset, lSize;
  Char strFileAndPath[128];
  
  if(f->part[0].id == 0xffff)
  {
    // Write file direct to card
    StrPrintF(strFileAndPath, "%s/%s", frodoDir,	f->name);

    // Search file in list of pdb's and .d64 files
    for(i=0; i<nitem; ++i)
    {
      if(StrCompare(f->name, file[item[i].fileindex].name) == 0)
      {
        ErrTry
        {
          err = VFSFileOpen(file[item[i].fileindex].card, strFileAndPath, vfsModeWrite | vfsModeTruncate, &fr);
          if(err) 
          	ErrThrow(err);
          
          // Loop all parts of file
          lRemainingSize = swap32(f->size);
          lCurrPart = 0;
          while(lRemainingSize > 0)
          {
            // Get size of current part
            if(lRemainingSize > swap32(f->partSize))
              lSize = swap32(f->partSize);
            else 
              lSize = lRemainingSize;
    
            // Write data of part
            err = VFSFileWrite(fr, lSize, (void *) swap32((UInt32)f->part[lCurrPart].p), &lNumBytesIO);
            if(err) 
            	ErrThrow(err);
            
            lRemainingSize = lRemainingSize - swap32(f->partSize);
            ++lCurrPart;
          }
        }
        ErrCatch(err)
        {
          DmResID lID;
          
          switch(err)
          {
            case errNone:             lID = StringErrNone; break;
            case expErrCardReadOnly:  lID = StringExpErrCardReadOnly; break;
            case expErrNotOpen:       lID = StringExpErrNotOpen; break;
            case vfsErrBadName:       lID = StringVfsErrBadName; break;
            case vfsErrFileNotFound:  lID = StringVfsErrFileNotFound; break;
            case vfsErrFilePermissionDenied: lID = StringVfsErrFilePermissionDenied; break;
            case vfsErrVolumeBadRef:  lID = StringVfsErrVolumeBadRef; break;
            default:                  lID = StringSaveFileErrDefault; break;
          }
          ShowErrorRes(lID);
        } ErrEndCatch
      
        // Close file
        if(fr != NULL)
          VFSFileClose(fr);
        
        break;
      }
    }
  }
  else
  {
    // Search file in list of pdb's and .d64 files
    for(i=0; i<nitem; ++i)
    {
      if(StrCompare(f->name, &(item[i].name[2])) == 0 && item[i].index != 0xffff)
      {
        // Found pdb in which the .d64 file is inside
        StrPrintF(strFileAndPath, "%s/%s", frodoDir,	file[item[i].fileindex].name);
        ErrTry
        {
          err = VFSFileOpen(file[item[i].fileindex].card, strFileAndPath, vfsModeReadWrite, &fr);
          if(err) 
          	ErrThrow(err);
          
          // Seek to position with number of records
          err = VFSFileSeek(fr, vfsOriginBeginning, 0x4c);
          if(err) 
          	ErrThrow(err);
          err = VFSFileRead(fr, 2, &lNumRecs, &lNumBytesIO);
          if(err) 
          	ErrThrow(err);

          // Loop all parts of file
          lRemainingSize = swap32(f->size);
          lCurrPart = 0;
          while(lRemainingSize > 0)
          {
            lIndex = f->part[lCurrPart].id;
            // Seek to offset of part
            err = VFSFileSeek(fr, vfsOriginBeginning, 0x4e + 8 * lIndex);
            if(err) 
            	ErrThrow(err);
            
            // Read start offset
            err = VFSFileRead(fr, 4, &lStartOffset, &lNumBytesIO);
            if(err) 
            	ErrThrow(err);
            
            // Read end offset or file len
            if(lIndex == lNumRecs - 1)
              err = VFSFileSize(fr, &lEndOffset);
            else
            {
              err = VFSFileSeek(fr, vfsOriginCurrent, 4);
              if(err) 
              	ErrThrow(err);
              err = VFSFileRead(fr, 4, &lEndOffset, &lNumBytesIO);
            }
            if(err) 
            	ErrThrow(err);
            lSize = lEndOffset - lStartOffset;

            // Move to position of part
            err = VFSFileSeek(fr, vfsOriginBeginning, lStartOffset + 52); // 52: Header of each part
            if(err) 
            	ErrThrow(err);
            
            // Write data of part
            err = VFSFileWrite(fr, lSize - 52, (void *) swap32((UInt32)f->part[lCurrPart].p), &lNumBytesIO);
            if(err) 
            	ErrThrow(err);
            
            lRemainingSize = lRemainingSize - swap32(f->partSize);
            ++lCurrPart;
          }
        }
        ErrCatch(err)
        {
          if(err)
            ShowErrorRes(StringErrWriteToSD);
        } ErrEndCatch

        // Close file
        if(fr != NULL)
          VFSFileClose(fr);
        
        break;
      }
    }
  }
}


//***********************************************************
// Procedure: SaveSnapshot
//
// Parameter: name  Name of snapshot
//            i     ID of record in pdb file
//
// Returnval: -
//
// Description: Save a snapshot to main db
//***********************************************************
void SaveSnapshot(char *name, Int32 i)
{
	MemHandle h1, h2;
	MemPtr p1, p2;
	Boolean ok = false;
	UInt16 idx, idx2;
	UInt16 n;
	char copiedName[32];
	UInt32 lOrgSize1, lOrgSize2;
	
	armletData->cmd &= ~(CMD68K_SAVE);
	if (!mainDBRef) 
		return;
	
	// Copy name so we access valid 32 chars during DmWrite
	MemSet(copiedName, 32, 0);
	StrNCopy(copiedName, name, 31);
	
	n = DmNumRecords(mainDBRef);
	idx = i;
	if (i>=0 && i<n) 
	{
		// Overwrite
		h1 = DmGetRecord(mainDBRef, idx);
		if (h1)
		{
			h2 = DmGetRecord(mainDBRef, idx+1);
			if(h2)
			{
  			lOrgSize1 = MemHandleSize(h1);
  			lOrgSize2 = MemHandleSize(h2);
  			if(lOrgSize1 < SNAPSHOT_SIZE_1 + 36 || lOrgSize2 < SNAPSHOT_SIZE_2 + 36 + 0x30)
  			{
  			  // Snapshot from previous version -> resize
  			  DmReleaseRecord(mainDBRef, idx, ok);
  			  h1 = DmResizeRecord(mainDBRef, idx, SNAPSHOT_SIZE_1 + 36);
  			  DmReleaseRecord(mainDBRef, idx+1, ok);
  			  h2 = DmResizeRecord(mainDBRef, idx+1, SNAPSHOT_SIZE_2 + 36 + 0x30);
  			}
      }
		}
	} 
	else 
	{
		// New
		idx = dmMaxRecordIndex;
		h1 = DmNewRecord(mainDBRef, &idx, SNAPSHOT_SIZE_1+36);
		if (h1) 
		{
			idx2 = idx+1;
			h2 = DmNewRecord(mainDBRef, &idx2, SNAPSHOT_SIZE_2+36+0x30);
		}
	}
	if (h1) 
	{
    if (h2) 
    {
  		p1 = MemHandleLock(h1);
  		if (p1) 
  		{
  			p2 = MemHandleLock(h2);
  			if (p2) 
  			{
  				DmWrite(p1, 0, "SNAP", 4);
  				DmWrite(p1, 4, copiedName, 32);
  				DmWrite(p1, 36, (MemPtr)swap32((UInt32)armletData->snapshot1), SNAPSHOT_SIZE_1);
  				DmWrite(p2, 0, "SNP2", 4);
  				DmWrite(p2, 4, copiedName, 32);
  				DmWrite(p2, 36, (MemPtr)swap32((UInt32)armletData->snapshot2), SNAPSHOT_SIZE_2);
  				DmWrite(p2, 36+SNAPSHOT_SIZE_2, prefs->DrivePath, 32);
  				DmWrite(p2, 36+SNAPSHOT_SIZE_2+32, &(prefs->DriveCard), 4);
  				DmWrite(p2, 36+SNAPSHOT_SIZE_2+36, &(prefs->DriveIndex), 4);
  				
          ok = true;
  				MemHandleUnlock(h2);
  			}
  			MemHandleUnlock(h1);
  		}
  		DmReleaseRecord(mainDBRef, idx+1, ok);
    }
    DmReleaseRecord(mainDBRef, idx, ok);
	}
	if (ok && idx!=0) // ignore default snapshot with index  0
	{ 
		lastIndex = idx;
		StrNCopy(lastName, name, 31);
		lastName[31] = '\0';
	}
	if (!ok)
	{
		ShowErrorRes(StringErrFailedToWriteSnap);
	}
  else
		if(bSyncRequired) 
		  DmSyncDatabase(mainDBRef); // Hope to fix bug with not stored prefs
}


//***********************************************************
// Procedure: LoadSnapshotFromSD
//
// Parameter: card  Card number to save
//            name  Name of snapshot
//
// Returnval: Errorcode
//
// Description: Load a snapshot to SD-Card
//***********************************************************
Err LoadSnapshotFromSD(UInt16 card, char *name)
{
  Char strFileAndPath[128];
  Err err = errNone;
  FileRef fr = NULL;
  UInt32 lNumBytesIO;

  StrPrintF(strFileAndPath, "%s/%s", frodoDir, name);

  err = VFSFileOpen(card, strFileAndPath, vfsModeRead, &fr);
  if(err) 
  	return err;

  // Load part one
	err = VFSFileRead(fr, SNAPSHOT_SIZE_1, (MemPtr)swap32((UInt32)armletData->snapshot1), &lNumBytesIO);
  if(err == errNone)
  {
    // Load part two
  	err = VFSFileRead(fr, SNAPSHOT_SIZE_2, (MemPtr)swap32((UInt32)armletData->snapshot2), &lNumBytesIO);
    if(err == errNone)
    {
      // Load current disk
    	err = VFSFileRead(fr, 32, (MemPtr) prefs->DrivePath, &lNumBytesIO);
      if(err == errNone)
      {
      	err = VFSFileRead(fr, 4, (MemPtr) &(prefs->DriveCard), &lNumBytesIO);
        if(err == errNone)
        {
        	err = VFSFileRead(fr, 4, (MemPtr) &(prefs->DriveIndex), &lNumBytesIO);
          if(err == errNone)
          {
            // Snapshot loaded -> store name and card num
            if(StrLen(name) - 4 < 31)
            {
  						StrNCopy(lastName, name, StrLen(name) - 4);
              lastName[StrLen(name) - 4]='\0';
            }
            else
            {
  						StrNCopy(lastName, name, 31);
  						lastName[31]='\0';
            }
						lastCard = card;
            
            armletData->cmd |= CMD68K_LOAD;
        		if(prefs->DrivePath[0] != 0)
        		  doFileLoadAfterSnap = true;
          }
        }
      }
    }
  }

  VFSFileClose(fr);
  return err;
}


//***********************************************************
// Procedure: SaveSnapshot2SD
//
// Parameter: name  Name of snapshot
//            card  Card number to save
//
// Returnval: -
//
// Description: Save a snapshot to SD-Card
//***********************************************************
void SaveSnapshot2SD(char *name, UInt16 card)
{
  Char strFileAndPath[128];
  Err err;
  FileRef fr = NULL;
  UInt32 lNumBytesIO;

  StrPrintF(strFileAndPath, "%s/%s.snp", frodoDir, name);
  
  ErrTry
  {
    err = VFSFileCreate(card, strFileAndPath);
    err = VFSFileOpen(card, strFileAndPath, vfsModeWrite | vfsModeTruncate, &fr);
    if(err) 
    	ErrThrow(err);

    // Write first part
    err = VFSFileWrite(fr, SNAPSHOT_SIZE_1, (MemPtr)swap32((UInt32)armletData->snapshot1), &lNumBytesIO);
    if(err) 
    	ErrThrow(err);

    // Write second part
    err = VFSFileWrite(fr, SNAPSHOT_SIZE_2, (MemPtr)swap32((UInt32)armletData->snapshot2), &lNumBytesIO);
    if(err) 
    	ErrThrow(err);

    // Write info about current disk
    err = VFSFileWrite(fr, 32, (MemPtr) prefs->DrivePath, &lNumBytesIO);
    if(err) 
    	ErrThrow(err);
    err = VFSFileWrite(fr, 4, (MemPtr) &(prefs->DriveCard), &lNumBytesIO);
    if(err) 
    	ErrThrow(err);
    err = VFSFileWrite(fr, 4, (MemPtr) &(prefs->DriveIndex), &lNumBytesIO);
    if(err) 
    	ErrThrow(err);
  }
  ErrCatch(err)
  {
    DmResID lID;
    
    switch(err)
    {
      case errNone:             lID = StringErrNone; break;
      case expErrCardReadOnly:  lID = StringExpErrCardReadOnly; break;
      case expErrNotOpen:       lID = StringExpErrNotOpen; break;
      case vfsErrBadName:       lID = StringVfsErrBadName; break;
      case vfsErrFileNotFound:  lID = StringVfsErrFileNotFound; break;
      case vfsErrFilePermissionDenied: lID = StringVfsErrFilePermissionDenied; break;
      case vfsErrVolumeBadRef:  lID = StringVfsErrVolumeBadRef; break;
      default:                  lID = StringSaveFileErrDefault; break;
    }
    ShowErrorRes(lID);
  } ErrEndCatch

  // Close file
  if(fr != NULL)
    VFSFileClose(fr);

  // Store last used snapshot name
	StrNCopy(lastName, name, 31);
	lastName[31] = '\0';
}


//***********************************************************
// Procedure: SavePrefsPalm
//
// Parameter: -
//
// Returnval: Errorcode or 0 if successful
//
// Description: Save preferences for Palm
//***********************************************************
UInt16 SavePrefsPalm(void)
{
	if (!prefsPalm) 
	{
		ShowErrorRes(StringErrNoPrefsToSave);
		return 5;
	}

  PrefSetAppPreferences(appSignature, IDX_PALMPREFS, 1, prefsPalm, sizeof(struct PreferencesPalm), false);

	return 0;
}


//***********************************************************
// Procedure: SavePrefs
//
// Parameter: -
//
// Returnval: Errorcode or 0 if successful
//
// Description: Save preferences
//***********************************************************
UInt16 SavePrefs(void)
{
	if (!prefs) 
	{
		ShowErrorRes(StringErrNoPrefsToSave);
		return 5;
	}

  PrefSetAppPreferences(appSignature, IDX_PREFS, 1, prefs, sizeof(struct PrefsState), false);

	return 0;
}


//*******************************************************************************************
//*******************************************************************************************
// List of files
//*******************************************************************************************
//*******************************************************************************************

//***********************************************************
// Procedure: FreeItems
//
// Parameter: -
//
// Returnval: -
//
// Description: Delete internal list of available disks. 
//***********************************************************
void FreeItems(void)
{
	if(file) 
	{
		if (fileH) 
			MemHandleUnlock(fileH);
		file = NULL;
	}
	if (fileH) 
	{
		MemHandleFree(fileH);
		fileH = NULL;
	}
	nfile = maxfile = 0;

	if (item) 
	{
		if (itemH) 
			MemHandleUnlock(itemH);
		item = NULL;
	}
	if (itemH) 
	{
		MemHandleFree(itemH);
		itemH = NULL;
	}
	nitem = maxitem = 0;
	
	if (itemList) 
	{
		if (itemListH) 
			MemHandleUnlock(itemListH);
		itemList = NULL;
	}
	if (itemListH) 
	{
		MemHandleFree(itemListH);
		itemListH = NULL;
	}
	nitemList = maxitemList = 0;
}


//***********************************************************
// Procedure: Scan
//
// Parameter: dbf   Pointer to file definition (see PalmOSGUI_dbfile.h)
//            card  Number of SD card
//            fnam  Name of file
//
// Returnval: Errorcode or 0 if successful
//
// Description: Read list of disks of file. 
//***********************************************************
Err Scan(DbFile *dbf, UInt16 card, char *fnam)
{
	UInt16 i, n;
	UInt32 type, app;
	MemHandle h;
	MemPtr p;
	RectangleType r={{20,70},{120,20}};
	char name[32] = "";
	FileRef index = NULL;
	Err err;
	char s[37];
  Char strFileAndPath[128];
  char strMessage[128];
  
  if (!dbf) 
		return 0;

	DbFileInfo(dbf, &type, &app, &n);

	err = AddFile(card, fnam);
	if (err) 
		return err;
	if (fnam && card != 0xffff) 
	{
		StrPrintF(strFileAndPath, "%s/%s", frodoDir, fnam);
		StrCopy(strFileAndPath + StrLen(strFileAndPath) - 4, ".idx");
		err = VFSFileOpen(card, strFileAndPath, vfsModeWrite | vfsModeCreate | vfsModeTruncate,	&index);
		if (err) 
		{
			index = 0;
			ShowErrorResLong(StringErrFailedToCreateIndex, err);
		}
	}

	// Scan a file for programs
	for (i=0; i<n; i++) 
	{
		err = DbFileGetRecord(dbf, i, &h);

		WinEraseRectangle(&r, 0);
		WinDrawRectangleFrame(simpleFrame, &r);
		StrPrintF(strMessage, "Db record %d/%d", i, n);
		WinDrawTruncChars(strMessage, StrLen(strMessage),	r.topLeft.x+10, r.topLeft.y+5, r.extent.x-20);

		if (!h) 
			continue;
		
		p = MemHandleLock(h);
		if (p) 
		{
			StrNCopy(s, p, 36);
			s[35] = '\0'; 
			if (!StrNCompare(p, "SNAP", 4) || !StrNCompare(p, "SPLT", 4)) 
			{
				err = AddItem(i, s+4);
				if (err) 
				{
					MemHandleUnlock(h);
					DbFileReleaseRecord(dbf, i, h, false);
					if (index) 
						VFSFileClose(index);
					return err;
				}
				if (dbf->external) 
				{
				  StrCopy(name, item[nitem-1].name+2);
				} 
				else 
				{
			    StrCopy(name, item[nitem-1].name);
				}
				if (index) 
				{
					VFSFileWrite(index, 2, &i, NULL);
					VFSFileWrite(index, 32, name, NULL);
				}
			} 
			else 
			{
				if (!StrCompare(name, "")) 
				{
					StrPrintF(strMessage, "Expected a SNAP @ %d, not %s",	i, s);
					ShowError(strMessage);
				} 
				else if (StrNCompare(name, p+4, 32)) 
				{
					StrPrintF(strMessage, "Expected a %s part @ %d, not %s",	name, i, s);
					ShowError(strMessage);
				}
			}
			MemHandleUnlock(h);
		}
		DbFileReleaseRecord(dbf, i, h, false);
	} // end for each record

	if (index) 
		VFSFileClose(index);

	return 0;
}


//***********************************************************
// Procedure: AddFile
//
// Parameter: card  Number of SD card
//            fnam  Name of file
//
// Returnval: Errorcode or 0 if successful
//
// Description: Add file to internal list
//***********************************************************
Err AddFile(UInt16 card, char *fnam)
{
	Err err;

	err = Grow(&fileH, (MemPtr *) &file, sizeof(struct t_file), nfile+1, &maxfile);
	if (err) 
		return err;

	file[nfile].card = card;
	MemSet(file[nfile].name, sizeof(file[nfile].name), 0x0);
	if (fnam) 
		StrNCopy(file[nfile].name, fnam, sizeof(file[nfile].name)-1);

	nfile++;
	return 0;
}


//***********************************************************
// Procedure: AddItem
//
// Parameter: idx   Index of record in pdb of 0xffff if .d64 file
//            fnam  Name of file
//
// Returnval: Errorcode or 0 if successful
//
// Description: Add file to internal list of available disks
//***********************************************************
Err AddItem(UInt16 idx, char *fnam)
{
	Err err;

	err = Grow(&itemH, (MemPtr *) &item, sizeof(struct t_item), nitem+1, &maxitem);
	if (err) 
		return err;

	item[nitem].fileindex = nfile-1;
	item[nitem].index = idx;
	MemSet(item[nitem].name, sizeof(item[nitem].name), 0x0);
	if (fnam) 
	{
		if (file[nfile-1].card == 0xffff) 
		{
			StrNCopy(item[nitem].name, fnam, sizeof(item[nitem].name)-3);
		} 
		else 
		{
			StrCopy(item[nitem].name, "X ");
			StrNCopy(item[nitem].name+2, fnam, sizeof(item[nitem].name)-3);
		}
	}
	nitem++;

	err = Grow(&itemListH, (MemPtr *) &itemList, sizeof(char *), nitemList+1, &maxitemList);
	if (err) 
		return err;
	nitemList++;

	return 0;
}


//***********************************************************
// Procedure: Grow
//
// Parameter: h     Handle of current memory
//            p     Pointer of locked memory
//            size  Size of one element
//            n     Number of elements
//            max   Number of max. elements
//
// Returnval: Errorcode or 0 if successful
//
// Description: Increase memory for any structure to hold more elements.
//***********************************************************
Err Grow(MemHandle *h, MemPtr *p, UInt16 size, UInt16 n, UInt16 *max)
{
	Err e;
  
	if (!*h) 
	{
		*max = 12;
		*h = MemHandleNew(12L * size);
		if (!(*h)) 
			return memErrNotEnoughSpace;
	} 
	else if (n > *max) 
	{
    // Check if we reached limit of 64K
    if((((UInt32)*max) * 2 * size) > 63000)
      return memErrNotEnoughSpace; // Don't grow but current memory remains valid
		*max *= 2;
		if (*p) 
		{
			MemHandleUnlock(*h);
			*p = NULL;
		}
		e = MemHandleResize(*h, ((UInt32)*max) * size);
		if (e) 
		{
			// Failure on resize, reference manual says nothing if previous allocated memory remains valid
			// try to lock previous memory handle
			*max /= 2;
			*p = MemHandleLock(*h);
			return e;
		}
	} 
	else 
	{
		// No change
		return 0;
	}

	*p = MemHandleLock(*h);
	if (!*p) 
	{
		MemHandleFree(*h);
		*h = NULL;
		return memErrNotEnoughSpace;
	}
	
	return 0;
}


//***********************************************************
// Procedure: ScanIndex
//
// Parameter: f     File reference handle
//            card  Number of SD card
//            fnam  Name of index-file
//
// Returnval: Errorcode or 0 if successful
//
// Description: Read index file with list disks. 
//***********************************************************
Err ScanIndex(FileRef f, UInt16 card, char *fnam)
{
	UInt16 idx;
	char name[32];
	Err err;

	err = AddFile(card, fnam);
	if (err) 
		return err;
	while (!VFSFileEOF(f)) 
	{
		err=VFSRead(f, 2, &idx);
		if (err) 
			return err;
		err=VFSRead(f, 32, name);
		if (err) 
			return err;
		err = AddItem(idx, name);
	}
	return 0;
}


//***********************************************************
// Procedure: ScanFiles
//
// Parameter: f   Handle to form to fill listbox
//
// Returnval: -
//
// Description: Create list of files
//***********************************************************
void ScanFiles(FormPtr f)
{
	UInt16 i;
	ListPtr l;
	char path[128], fnam[100], fnam2[100];
	UInt16 vol;
	UInt32 volIter, fIter;
	Err err;
	FileRef dir, file;
	FileInfoType info;
	DbFile dbf;
		
	// Just in case
	FreeItems();

	if(!bLaunchedFromSD)
  {
  	DbFileFromDbRef(mainDBRef, &dbf);
  	err = Scan(&dbf, 0xffff, NULL);
  	if (err) 
  	{
  		ShowErrorResLong(StringErrNoMemoryForScan, err);
  		return;
  	}
  }
  
	// For each expansion card
	volIter = vfsIteratorStart;
	while (volIter!=vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&vol, &volIter);
		if (err) 
			continue;
	
		err = VFSFileOpen(vol, frodoDir, vfsModeRead, &dir);
		if (err) 
			continue;

		info.nameP = fnam;
		info.nameBufLen = sizeof(fnam);

		// For each file in /Palm/Programs/Frodo
		fIter = expIteratorStart;
		while (fIter != expIteratorStop) 
		{
			err = VFSDirEntryEnumerate(dir, &fIter, &info);
			if (err) 
				continue;
			
			if (StrLen(fnam)<4) 
				continue;
			if (!StrCompare(fnam+StrLen(fnam)-4, ".pdb")) 
			{
				// Try to find cached index
				StrCopy(fnam2, fnam);
				StrCopy(fnam2+StrLen(fnam2)-4, ".idx");
				StrPrintF(path, "%s/%s", frodoDir, fnam2);
				err=VFSFileOpen(vol, path, vfsModeRead, &file);
				if (!err)  
				{
					err = ScanIndex(file, vol, fnam);
					if (err) 
					{
            ShowErrorResLong(StringErrNoMemoryForScan, err);
						break;
					}
					continue;
				}

        // No index file -> scan pdb
				StrPrintF(path, "%s/%s", frodoDir, fnam);
				err = VFSFileOpen(vol, path, vfsModeRead, &file);
				if (err) 
					continue;

				DbFileFromFileRef(file, &dbf);
				err = Scan(&dbf, vol, fnam);
				VFSFileClose(file);
				if (err) 
				{
          ShowErrorResLong(StringErrNoMemoryForScan, err);
					break;
				}
			} 
			else if (!StrCaselessCompare(fnam+StrLen(fnam)-4, ".d64") ||
			         !StrCaselessCompare(fnam+StrLen(fnam)-4, ".t64") ||
			         !StrCaselessCompare(fnam+StrLen(fnam)-4, ".p00") ||
			         !StrCaselessCompare(fnam+StrLen(fnam)-4, ".snp")) 
			{
				// Add this one file (.d64, .t64, .p00 or .snp)
				err = AddFile(vol, fnam);
				if (err) 
				{
					ShowErrorResLong(StringErrNoMemoryForScan, err);
					break;
				}
				err = AddItem(0xffff, fnam);
				if (err) 
				{
					ShowErrorResLong(StringErrNoMemoryForScan, err);
					break;
				}
			}
		}
		VFSFileClose(dir);
	}

	if (itemList) 
	{
		Int16 lidx = 0;
    for (i=0; i<nitem && i<nitemList; i++)
    {
	  	itemList[i] = item[i].name;
			if (item[i].index == (UInt16)lastIndex && lastIndex>0) 
				lidx=i;
		}
    if(0 != f)
    {
      i = FrmGetObjectIndex(f, FileFileList);
      if (i!=frmInvalidObjectId) 
      {
    		l = (ListType *) FrmGetObjectPtr(f, i);
    		LstSetListChoices(l, itemList, nitemList);
  	  	LstDrawList(l);
				if (lidx>0) LstSetSelection(l,lidx);
      }
    }
	}
}


//***********************************************************
// Procedure: LoadFile
//
// Parameter: card  Number of card
//            name  Filename
//
// Returnval: Errorcode or 0 if successful
//
// Description: Load a .d64 or .t64 file direct from expansion card
//***********************************************************
Err LoadFile(UInt16 card, char *name)
{
  FileRef lFile = 0;
  Err error = errNone;
  UInt32 lSize, lDummy;
  UInt32 lNumParts = 0, lPartSize;
  UInt32 i, j;
  Char strFileAndPath[128];
  UInt32 PARTSIZE = 61440;
  UInt32 p00Size = 0;

	if (!StrNCaselessCompare(name+StrLen(name)-4, ".snp", 4))
  {
    // Load snapshot from SD-Card
    return LoadSnapshotFromSD(card, name);
  }
  
  // Clear actual content of drive
  Close((FILE *) swap32((UInt32)armletData->disk));
	armletData->disk = NULL;
	MemSet(&disk, sizeof(FILE), 0x0);

  ErrTry
  {
    // Open file for reading
    StrCopy(strFileAndPath, frodoDir);
    StrCat(strFileAndPath, "/");
    StrCat(strFileAndPath, name);
    error = VFSFileOpen(card, strFileAndPath, vfsModeRead, &lFile);
    if(error != errNone)
      ErrThrow(error);
    
    // Get size of file
    error = VFSFileSize(lFile, &lSize);
    if(error != errNone)
      ErrThrow(error);
		if (!StrNCaselessCompare(name+StrLen(name)-4, ".p00", 4))
		{
			p00Size = lSize;
			lSize += 0x80 - 0x1c; // Header of .t64 minus header of p00
		}
		
    // Get number of memory blocks
    if(lSize%PARTSIZE == 0)
      lNumParts = lSize/PARTSIZE;
    else
      lNumParts = lSize/PARTSIZE + 1;

    // Store name of file
    StrNCopy(disk.name, name, 31);
    disk.name[31] = 0;
    
    // Store properties of file
		if (!StrNCaselessCompare(disk.name + StrLen(disk.name)-4, ".d64", 4))
      disk.type = FILETYPE_D64;
    else
      disk.type = FILETYPE_T64;
    disk.size = lSize;
    disk.partSize = PARTSIZE;
    disk.offset = 0;
    disk.partOffset = 0;
    disk.currentPart = 0;
    disk.modified = 0;
    disk.location = 1;
    
    // Loop all parts
    for(i=0; i<lNumParts; ++i)
    {
      // Get memory for current part
      if(i+1 < lNumParts)
        lPartSize = PARTSIZE;
      else
        lPartSize = lSize%PARTSIZE;
      disk.part[i].handle = (uint32) MemHandleNew(lPartSize);
      if(disk.part[i].handle == 0)
        ErrThrow(memErrNotEnoughSpace);
      disk.part[i].p = MemHandleLock((MemHandle) disk.part[i].handle);
      disk.part[i].dbf = 0;
      disk.part[i].id = 0xffff;

      // Load part to memory
      if(p00Size)
      {
				UInt32 StartAdr, EndAdr;
				
				// p00 never larger than 61440 -> fit into one block
     		error = VFSFileRead(lFile, p00Size, disk.part[i].p + 0x80 - 0x1c, &lDummy); // Now Data in correct place
				
				// Create header of .t64
				for(j=0; j<0x80 - 0x1c; ++j)
					disk.part[i].p[j] = 0;
				StrCopy(disk.part[i].p, "C64 tape image file");
				disk.part[i].p[0x22] = 1; // Only one file
				for(j=0; j<0x10; ++j)
				{
					disk.part[i].p[0x28 + j] = disk.part[i].p[0x80 - 0x14 + j]; // Name of disk is name of file
					if(disk.part[i].p[0x28 + j] == 0)
					  disk.part[i].p[0x28 + j] = 0x20; // Use Space instead of 0
				}
				disk.part[i].p[0x40] = 1; // Valid file
				disk.part[i].p[0x42] = disk.part[i].p[0x80 - 0x2]; // Start address low
				disk.part[i].p[0x43] = disk.part[i].p[0x80 - 0x1]; // Start address high
				StartAdr = disk.part[i].p[0x42] | (disk.part[i].p[0x43] << 8);
				EndAdr = StartAdr + (p00Size - 0x1c);
				disk.part[i].p[0x44] = EndAdr & 0xff; // End address low
				disk.part[i].p[0x45] = EndAdr >> 8;   // End address high
				disk.part[i].p[0x48] = 0x80; // Offset in file
				for(j=0; j<0x10; ++j)
					disk.part[i].p[0x50 + j] = disk.part[i].p[0x28 + j]; // Name of file
				disk.part[i].p[0x60] = 0; // No more files
      }
      else
      	error = VFSFileRead(lFile, lPartSize, disk.part[i].p, &lDummy);
      if(error != errNone)
        ErrThrow(error);
    }       
  }
  ErrCatch(inErr)
  {
    error = inErr;
  } ErrEndCatch
  
  // Close file
  if(lFile)
    VFSFileClose(lFile);

  if(error == errNone)
  {
    // Rearrange all data to little endian for use in ARM code
  	disk.size = swap32(disk.size);
		disk.partSize = swap32(disk.partSize);
		for (i=0; i<lNumParts; ++i)
		{
			disk.part[i].p = (UInt8 *) swap32((UInt32) disk.part[i].p);
		}
		armletData->disk = (FILE *) swap32((UInt32)&disk);
		if (!StrNCaselessCompare(disk.name+StrLen(disk.name)-4, ".d64", 4))
			prefs->DriveType = swap32(DRVTYPE_D64);
		else
		{
			prefs->DriveType = swap32(DRVTYPE_T64);
			if(prefs->Emul1541Proc)
			{
			  // No emulation if t64 or p00 file
			  prefs->Emul1541Proc = 0;
			  ShowInfoRes(StringWarn1541Off);
				SavePrefs();
			  SavePrefsPalm();
			}
		}
		MemMove(prefs->DrivePath, disk.name, 32);
    prefs->DriveCard = card;
    prefs->DriveIndex = 0xffff;
    
    // Send signal to C64 that new disk attached
		armletData->cmd |= CMD68K_SETPREFS;
	}
	else
	{
	  // Failed to load disk
		armletData->cmd &= ~(CMD68K_SETPREFS);
		//ShowErrorResStr(StringErrFailedToLoadDiskImage, name);
		MemSet(prefs->DrivePath, 32, 0);
    prefs->DriveCard = 0;
    prefs->DriveIndex = 0xffff;
    prefs->DriveType = 0;
	}
  
  return error;    
}


//***********************************************************
// Procedure: LoadSelectedFile
//
// Parameter: i   Index of file in list
//
// Returnval: -
//
// Description: Load a snapshot, d64 or t64 from any location
//***********************************************************
void LoadSelectedFile(UInt16 i)
{
  LoadSpecifiedFile(file[item[i].fileindex].card, item[i].index, file[item[i].fileindex].name);
}


//***********************************************************
// Procedure: LoadSpecifiedFile
//
// Parameter: card    Number of card of file (0xFFFF if internal)
//            index   Index if file is in pdb
//            name    Name of file
//
// Returnval: -
//
// Description: Load a snapshot, d64 or t64 from any location
//***********************************************************
void LoadSpecifiedFile(UInt16 card, UInt16 index, char *name)
{
  Char strFileAndPath[128];
	FileRef fr;
	DbFile dbf;
	Err err;

	if (card == 0xffff) 
	{
		// Internal pdb
		DbFileFromDbRef(mainDBRef, &dbf);
		Load(&dbf, index, card);
	} 
	else if (index == 0xffff) 
	{
		// external file
		LoadFile(card, name);
	} 
	else 
	{
		// external pdb
		StrPrintF(strFileAndPath, "%s/%s", frodoDir, name);
		err = VFSFileOpen(card, strFileAndPath, vfsModeRead, &fr);
		if (!err) 
		{
			DbFileFromFileRef(fr, &dbf);
			Load(&dbf, index, card);
		}
		VFSFileClose(fr);
	}
}


//***********************************************************
// Procedure: CheckValidCard
//
// Parameter: card    Number of card of file (0xFFFF if internal)
//            name    Name of file
//
// Returnval: Number of card with file
//
// Description: Check if file exists on card
//***********************************************************
UInt16 CheckValidCard(UInt16 card, char *name)
{
  FileRef lFile = 0;
  Char strFileAndPath[128];
  UInt16 lVolRefNum;
  UInt32 lVolIterator;

  // Open file for reading
  StrCopy(strFileAndPath, frodoDir);
  StrCat(strFileAndPath, "/");
  StrCat(strFileAndPath, name);
  if(VFSFileOpen(card, strFileAndPath, vfsModeRead, &lFile) != errNone)
  {
    // File not found on card -> search all cards
    lVolIterator = vfsIteratorStart;
    while(lVolIterator != vfsIteratorStop)
    {
      if(errNone == VFSVolumeEnumerate(&lVolRefNum, &lVolIterator))
      {
        if(VFSFileOpen(lVolRefNum, strFileAndPath, vfsModeRead, &lFile) == errNone)
        {
          // Found file on different card
          VFSFileClose(lFile);
          return lVolRefNum;
        }
      }
    }
    return 0;    
  }
  VFSFileClose(lFile);
  return card;
}


//***********************************************************
// Procedure: CheckFileExists
//
// Parameter: card    Number of card of file (0xFFFF if internal)
//            index   Index if file is in pdb
//            name    Name of file
//
// Returnval: True if file exists
//
// Description: Check if file exists
//***********************************************************
Boolean CheckFileExists(UInt16 card, UInt16 index, char *name)
{
	if (card == 0xffff) 
	{
		// Internal pdb
		return true;
	} 
	else //if (index == 0xffff) 
	{
		// external file or pdb -> check if card is correct
		card = CheckValidCard(card, name);
		if(card != 0)
		{
      prefs->DriveCard = card;		  
		  return true;
		}
    else
	  {
  		// File not found on any card
  		ShowErrorResStr(StringErrFailedToLoadDiskImage, name);
  		MemSet(prefs->DrivePath, 32, 0);
      prefs->DriveCard = 0;
      prefs->DriveIndex = 0xffff;
      prefs->DriveType = 0;
      return false;
	  }
	}  
}


//***********************************************************
// Procedure: Delete
//
// Parameter: dbf   Pointer to file definition (see PalmOSGUI_dbfile.h)
//            idx   ID of record in pdb file
//
// Returnval: -
//
// Description: Delete all parts of .d64/.t64 in pdb (not on sd card!)
//***********************************************************
void Delete(DbFile *dbf, UInt16 idx)
{
	MemHandle h;
	MemPtr p;
	UInt16 i, n;
	Boolean stop = false;
	DbFile mainFile;
	Err err;
	
	if (!dbf) 
	{
		dbf = &mainFile;
		DbFileFromDbRef(mainDBRef, dbf);
	}

	// Sanity checks
	if (dbf->external) 
		return;
	if (!dbf->u.db) 
		return;
	if (dbf->u.db==mainDBRef && idx<2) 
		return;
  n = DmNumRecords(mainDBRef);
	if (idx>=n) 
		return;
	if (lastIndex>0 && idx == (UInt16)lastIndex)
		lastIndex = INVALID_SNAPSHOT_INDEX;

	for (i=idx; i<n && !stop; i++) 
	{
		err = DbFileGetRecord(dbf, i, &h);
		if (h) 
		{
			p=MemHandleLock(h);
			if (p) 
			{
				if (i==idx) 
				{
					if (StrNCompare(p, "SNAP", 4) && StrNCompare(p, "SPLT", 4))
						stop = 1;
				} 
				else if (!StrNCompare(p, "SNAP", 4) || !StrNCompare(p, "SPLT", 4))
					stop = 1;
				MemHandleUnlock(h);
			} 
			else
				stop = true;
			DbFileReleaseRecord(dbf, i, h, false);
		} 
		else
			stop = true;
		if (!stop) 
			DmDeleteRecord(dbf->u.db, i);
	}
}


//***********************************************************
// Procedure: DeleteFile
//
// Parameter: i   Index of file in list
//
// Returnval: -
//
// Description: Delete a snapshot, d64 or t64
//***********************************************************
void DeleteFile(UInt16 i)
{
	DbFile dbf;
  Char strFileAndPath[128];

	if (i>=0)
	{
	  if(file[item[i].fileindex].card == 0xffff) 
  	{
  		DbFileFromDbRef(mainDBRef, &dbf);
  		Delete(&dbf, item[i].index);
    }
  	else if(item[i].index == 0xffff) 
  	{
      // Delete from SD-Card
      StrPrintF(strFileAndPath, "%s/%s", frodoDir, file[item[i].fileindex].name);
      VFSFileDelete(file[item[i].fileindex].card, strFileAndPath);
    }
	}
}


//***********************************************************
// Procedure: VFSRead
//
// Parameter: f   File reference handle
//            n   Number of bytes to read
//            vP  Pointer to buffer for data
//
// Returnval: Errorcode or 0 if successful
//
// Description: Read bytes from file on SD card. 
//***********************************************************
Err VFSRead(FileRef f, UInt32 n, void *vP)
{
	char *buf = (char *)vP;
	UInt32 r, tot = 0;
	Err err;

	while (tot<n) 
	{
		err = VFSFileRead(f, n-tot, buf+tot, &r);
		if (err) 
			return err;
		tot += r;
	}
	return 0;
}


//***********************************************************
// Procedure: SearchValidCard
//
// Parameter: -
//
// Returnval: Number of card (0xffff if no card available)
//
// Description: Search first card with frododir
//***********************************************************
UInt16 SearchValidCard()
{
  FileRef lFile = 0;
  UInt16 lVolRefNum;
  UInt32 lVolIterator;

  // Search all cards
  lVolIterator = vfsIteratorStart;
  while(lVolIterator != vfsIteratorStop)
  {
    if(errNone == VFSVolumeEnumerate(&lVolRefNum, &lVolIterator))
    {
      if(VFSFileOpen(lVolRefNum, frodoDir, vfsModeRead, &lFile) == errNone)
      {
        // Found frododir on card
        VFSFileClose(lFile);
        return lVolRefNum;
      }
    }
  }

  return 0xffff;    
}
